from cone.sql import get_session
from cone.sql import testing as sql_testing
from cone.tokens.api import TokenAPI
from cone.tokens.exceptions import TokenLockTimeViolation
from cone.tokens.exceptions import TokenNotExists
from cone.tokens.exceptions import TokenTimeRangeViolation
from cone.tokens.exceptions import TokenUsageCountExceeded
from cone.tokens.exceptions import TokenValueError
from cone.tokens.model import TokenRecord
from cone.tokens.model import TokenUsageRecord
from cone.tokens.tests import tokens_layer
from cone.ugm.testing import principals
from datetime import datetime 
from datetime import timedelta
from node.tests import NodeTestCase
import uuid


class TestTokenAPI(NodeTestCase):
    layer = tokens_layer

    def test_contextmanager(self):
        api = TokenAPI()
        with api:
            api.query_token(uuid.uuid4())

    @sql_testing.delete_table_records(TokenRecord)
    def test_get_token(self):
        request = self.layer.new_request()
        session = get_session(request)

        token = TokenRecord()
        token_uid = token.uid = uuid.uuid4()

        session.add(token)
        session.commit()

        api = TokenAPI(request)
        self.assertEqual(api.get_token(token_uid).uid, token_uid)

        invalid_uid = uuid.UUID('f9c98a6b-b773-4714-b965-98c7911e6236')
        with self.assertRaises(TokenNotExists) as arc:
            api.get_token(invalid_uid)
        self.assertEqual(
            str(arc.exception),
            'Token f9c98a6b-b773-4714-b965-98c7911e6236 not exists'
        )
        self.assertEqual(arc.exception.error_code, 1)

    @sql_testing.delete_table_records(TokenRecord)
    def test_query_token(self):
        request = self.layer.new_request()
        session = get_session(request)

        token = TokenRecord()
        token_uid = token.uid = uuid.uuid4()
        token_value = token.value = 'token value'

        session.add(token)
        session.commit()

        api = TokenAPI(request)
        self.assertEqual(api.query_token(''), None)
        self.assertEqual(api.query_token(token_value).uid, token_uid)

    @sql_testing.delete_table_records(TokenRecord)
    def test_token_values(self):
        request = self.layer.new_request()
        session = get_session(request)

        token = TokenRecord()
        token.uid = uuid.UUID('7217e39a-e39a-4e34-ae6d-30f127169962')
        token.value = '1'
        session.add(token)

        token = TokenRecord()
        token.uid = uuid.UUID('f9330598-eabd-415c-8049-beed35b9e9ab')
        token.value = '2'
        session.add(token)

        session.commit()

        api = TokenAPI(request)
        self.assertEqual(
            api.token_values([
                uuid.UUID('7217e39a-e39a-4e34-ae6d-30f127169962'),
                uuid.UUID('f9330598-eabd-415c-8049-beed35b9e9ab'),
                uuid.UUID('1258a13a-7694-4c42-bbd9-64464c558a06')
            ]),
            ['1', '2']
        )

    @sql_testing.delete_table_records(TokenRecord)
    @sql_testing.delete_table_records(TokenUsageRecord)
    def test_consume(self):
        request = self.layer.new_request()
        session = get_session(request)

        token = TokenRecord()
        token_uid = uuid.UUID('577989d4-1673-4639-a579-dd468b294713')
        token.uid = token_uid
        token.value = 'token value'
        last_used = token.last_used = datetime(2023, 1, 1)
        token.valid_from = None
        token.valid_to = None
        token.lock_time = 0
        token.usage_count = -1
        session.add(token)

        api = TokenAPI(request)
        self.assertTrue(api.consume(token_uid))
        self.assertNotEqual(last_used, token.last_used)
        self.assertEqual(token.usage_count, -1)

        usage_record = session.query(TokenUsageRecord).one()
        self.assertEqual(usage_record.token_uid, token_uid)
        self.assertIsInstance(usage_record.timestamp, datetime)
        self.assertEqual(usage_record.error_code, None)
        self.assertEqual(usage_record.user, None)

        token.usage_count = 1
        session.commit()
        self.assertTrue(api.consume(token_uid))
        self.assertEqual(token.usage_count, 0)
        self.assertEqual(session.query(TokenUsageRecord).count(), 2)

        with self.assertRaises(TokenUsageCountExceeded) as arc:
            api.consume(token_uid)
        self.assertEqual(
            str(arc.exception),
            'Token 577989d4-1673-4639-a579-dd468b294713 usage count exceeded'
        )
        self.assertEqual(arc.exception.error_code, 2)
        self.assertEqual(session.query(TokenUsageRecord).count(), 3)

        token.usage_count = -1
        token.lock_time = 120
        session.commit()
        with self.assertRaises(TokenLockTimeViolation) as arc:
            api.consume(token_uid)
        self.assertEqual(
            str(arc.exception),
            'Token 577989d4-1673-4639-a579-dd468b294713 is locked'
        )
        self.assertEqual(arc.exception.error_code, 3)
        self.assertEqual(session.query(TokenUsageRecord).count(), 4)

        token.lock_time = 0
        session.commit()
        self.assertTrue(api.consume(token_uid))
        self.assertEqual(session.query(TokenUsageRecord).count(), 5)

        token.valid_from = datetime.now() + timedelta(days=1)
        session.commit()
        with self.assertRaises(TokenTimeRangeViolation) as arc:
            api.consume(token_uid)
        self.assertEqual(
            str(arc.exception),
            'Token 577989d4-1673-4639-a579-dd468b294713 out of time range'
        )
        self.assertEqual(arc.exception.error_code, 4)
        self.assertEqual(session.query(TokenUsageRecord).count(), 6)

        token.valid_from = None
        token.valid_to = datetime.now() - timedelta(days=1)
        session.commit()
        with self.assertRaises(TokenTimeRangeViolation) as arc:
            api.consume(token_uid)
        self.assertEqual(
            str(arc.exception),
            'Token 577989d4-1673-4639-a579-dd468b294713 out of time range'
        )
        self.assertEqual(arc.exception.error_code, 4)
        self.assertEqual(session.query(TokenUsageRecord).count(), 7)

        token.valid_from = datetime.now() - timedelta(days=1)
        token.valid_to = datetime.now() + timedelta(days=1)
        session.commit()

        self.assertTrue(api.consume(token_uid))

        session.delete(token)
        with self.assertRaises(TokenNotExists) as arc:
            api.consume(token_uid)
        self.assertEqual(
            str(arc.exception),
            'Token 577989d4-1673-4639-a579-dd468b294713 not exists'
        )
        self.assertEqual(arc.exception.error_code, 1)
        self.assertEqual(session.query(TokenUsageRecord).count(), 8)

    @principals(
        users={
            'user_1': {}
        },
        groups={
            'group_1': {}
        },
        membership={
            'group_1': ['user_1']
        },
        roles={
            'user_1': ['manager']
    })
    @sql_testing.delete_table_records(TokenRecord)
    def test_add(self):
        request = self.layer.new_request()
        api = TokenAPI(request)
        token_uid = uuid.UUID('c27b6d86-8ac0-4261-9e62-151ff7e31ecb')
        with self.layer.authenticated('user_1'):
            api.add(token_uid)

        token = api.get_token(token_uid)
        self.assertEqual(token.value, str(token_uid))
        self.assertEqual(token.valid_from, None)
        self.assertEqual(token.valid_to, None)
        self.assertEqual(token.usage_count, 0)
        self.assertEqual(token.lock_time, 0)
        self.assertEqual(token.creator, 'user_1')
        self.assertTrue(token.created)
        self.assertTrue(token.modified)

        token_uid = uuid.UUID('6556f43e-b0ce-4c14-a0c5-40b8f2cdab3a')
        api.add(
            token_uid,
            value='token value',
            valid_from=datetime(2023, 9, 21),
            valid_to=datetime(2023, 9, 22),
            usage_count=1,
            lock_time=10
        )
        token = api.get_token(token_uid)
        self.assertEqual(token.value, 'token value')
        self.assertEqual(token.valid_from, datetime(2023, 9, 21))
        self.assertEqual(token.valid_to, datetime(2023, 9, 22))
        self.assertEqual(token.usage_count, 1)
        self.assertEqual(token.lock_time, 10)
        self.assertEqual(token.creator, None)

        with self.assertRaises(TokenValueError) as arc:
            api.add(
                token_uid,
                valid_from=datetime(2023, 9, 21),
                valid_to=datetime(2023, 9, 21),
            )
        self.assertEqual(
            str(arc.exception),
            'Token with uid 6556f43e-b0ce-4c14-a0c5-40b8f2cdab3a already exists'
        )
        self.assertEqual(arc.exception.error_code, 5)

        token_uid = uuid.UUID('9c9196f0-8b5b-42e7-b389-b13b267c9378')
        with self.assertRaises(TokenValueError) as arc:
            api.add(
                token_uid,
                valid_from=datetime(2023, 9, 21),
                valid_to=datetime(2023, 9, 21),
            )
        self.assertEqual(
            str(arc.exception),
            'valid_from must be before valid_to'
        )
        self.assertEqual(arc.exception.error_code, 5)

    @sql_testing.delete_table_records(TokenRecord)
    def test_delete(self):
        request = self.layer.new_request()
        session = get_session(request)

        token = TokenRecord()
        token_uid = token.uid = uuid.UUID('cc3ebacb-1fc1-42ea-bc24-051b60d60545')
        session.add(token)
        session.commit()

        api = TokenAPI(request)
        api.delete(token_uid)
        with self.assertRaises(TokenNotExists) as arc:
            api.delete(token_uid)
        self.assertEqual(
            str(arc.exception),
            'Token cc3ebacb-1fc1-42ea-bc24-051b60d60545 not exists'
        )
        self.assertEqual(arc.exception.error_code, 1)

    @sql_testing.delete_table_records(TokenRecord)
    def test_update(self):
        request = self.layer.new_request()
        session = get_session(request)

        token = TokenRecord()
        token_uid = token.uid = uuid.UUID('17b04001-4f42-4061-ad85-72de9ed8e287')
        token.value = 'value'
        session.add(token)

        token = TokenRecord()
        token.uid = uuid.UUID('c32e818c-af0c-4ee4-a855-0e50cad09e13')
        token.value = 'other value'
        session.add(token)
        session.commit()

        api = TokenAPI(request)
        with self.assertRaises(TokenValueError) as arc:
            api.update(token_uid, value='other value')
        self.assertEqual(
            str(arc.exception),
            'Given value already used by another token'
        )
        self.assertEqual(arc.exception.error_code, 5)

        with self.assertRaises(TokenValueError) as arc:
            api.update(
                token_uid,
                valid_from=datetime(2023, 9, 21),
                valid_to=datetime(2023, 9, 21)
            )
        self.assertEqual(
            str(arc.exception),
            'valid_from must be before valid_to'
        )
        self.assertEqual(arc.exception.error_code, 5)

        api.update(
            token_uid,
            value='new value',
            valid_from=datetime(2023, 9, 21),
            valid_to=datetime(2023, 9, 22),
            usage_count=10,
            lock_time=100
        )
        token = session\
            .query(TokenRecord)\
            .filter(TokenRecord.uid == token_uid)\
            .one()

        self.assertEqual(token.value, 'new value')
        self.assertEqual(token.valid_from, datetime(2023, 9, 21))
        self.assertEqual(token.valid_to, datetime(2023, 9, 22))
        self.assertEqual(token.usage_count, 10)
        self.assertEqual(token.lock_time, 100)



from cone.sql.testing import SQLLayer
import sys
import unittest


class TokensLayer(SQLLayer):

    def make_app(self):
        plugins = [
            'cone.ugm',
            'cone.sql',
            'cone.tokens'
        ]
        kw = dict()
        kw['cone.plugins'] = '\n'.join(plugins)
        kw['cone.tokens.config_file'] = '/tmp/tokens.json'
        super().make_app(**kw)


tokens_layer = TokensLayer()


def test_suite():
    from cone.tokens.tests import test_api
    from cone.tokens.tests import test_json_api
    from cone.tokens.tests import test_model
    from cone.tokens.tests import test_token

    suite = unittest.TestSuite()

    suite.addTest(unittest.findTestCases(test_api))
    suite.addTest(unittest.findTestCases(test_json_api))
    suite.addTest(unittest.findTestCases(test_model))
    suite.addTest(unittest.findTestCases(test_token))

    return suite


def run_tests():
    from zope.testrunner.runner import Runner

    runner = Runner(found_suites=[test_suite()])
    runner.run()
    sys.exit(int(runner.failed))


if __name__ == '__main__':
    run_tests()



from cone.app import get_root
from cone.tokens.browser.settings import TokenSettingsForm
from cone.tokens.settings import TokenSettings
from cone.tokens.settings import get_settings_node
from cone.tokens.settings import tokens_config
from cone.tokens.tests import tokens_layer
from cone.ugm import testing
from node.tests import NodeTestCase
from node.utils import UNSET
import json
import os


class TestSettings(NodeTestCase):
    layer = tokens_layer

    @testing.temp_directory
    def test_model_settings(self, tempdir):
        tokens_config.config_file = os.path.join(tempdir, 'tokens.json')

        # default - no config file exists
        settings = TokenSettings()
        attrs = settings.attrs
        self.assertEqual(attrs['morning']['start'], '08:00')
        self.assertEqual(attrs['morning']['end'], '12:00')
        self.assertEqual(attrs['afternoon']['start'], '12:00')
        self.assertEqual(attrs['afternoon']['end'], '18:00')
        self.assertEqual(attrs['today']['start'], '08:00')
        self.assertEqual(attrs['today']['end'], '18:00')
        self.assertEqual(attrs['default_locktime'], 0)
        self.assertEqual(attrs['default_usage_count'], 0)

        # config file exists
        with open(tokens_config.config_file, 'w') as f:
            attrs['morning']['start'] = '09:00'
            attrs['default_locktime'] = '2000'
            json.dump(attrs, f)

        attrs = settings.attrs
        self.assertEqual(attrs['morning']['start'], '09:00')
        self.assertEqual(attrs['default_locktime'], '2000')

    @testing.temp_directory
    def test_BrowserSettingsForm(self, tempdir):
        tokens_config.config_file = os.path.join(tempdir, 'tokens.json')
        model = get_settings_node(get_root())
        request = self.layer.new_request()

        tile = TokenSettingsForm()
        tile.model = model
        tile.request = request
        tile.prepare()

        form = tile.form
        self.assertEqual(form.keys(), [
            'morning',
            'afternoon',
            'today',
            'default_locktime',
            'default_usage_count',
            'save',
            'cancel',
            'came_from'
        ])
        self.assertEqual(form['morning'].keys(), ['start', 'end'])
        self.assertEqual(form['afternoon'].keys(), ['start', 'end'])
        self.assertEqual(form['today'].keys(), ['start', 'end'])

        # extract
        data = form.extract(request=request)
        morning = data.fetch('tokensettingsform.morning')
        self.assertEqual(morning.extracted['start'], UNSET)
        self.assertEqual(morning.extracted['end'], UNSET)
        afternoon = data.fetch('tokensettingsform.afternoon')
        self.assertEqual(afternoon.extracted['start'], UNSET)
        self.assertEqual(afternoon.extracted['end'], UNSET)
        today = data.fetch('tokensettingsform.today')
        self.assertEqual(today.extracted['start'], UNSET)
        self.assertEqual(today.extracted['end'], UNSET)
        self.assertEqual(
            data.fetch('tokensettingsform.default_locktime').extracted,
            UNSET
        )
        self.assertEqual(
            data.fetch('tokensettingsform.default_usage_count').extracted,
            UNSET
        )

        # save
        request.params['tokensettingsform.morning.start'] = '9:00'
        request.params['tokensettingsform.morning.end'] = '10:00'
        request.params['tokensettingsform.afternoon.start'] = '13:00'
        request.params['tokensettingsform.afternoon.end'] = '19:00'
        request.params['tokensettingsform.today.start'] = '9:00'
        request.params['tokensettingsform.today.end'] = '19:00'
        request.params['tokensettingsform.default_locktime'] = '2000'
        request.params['tokensettingsform.default_usage_count'] = '5'
        data = tile.form.extract(request=request)
        tile.save(model, data)

        with open(tokens_config.config_file) as f:
            form_data = json.load(f)
        self.assertEqual(form_data['morning']['start'], '09:00')
        self.assertEqual(form_data['morning']['end'], '10:00')
        self.assertEqual(form_data['afternoon']['start'], '13:00')
        self.assertEqual(form_data['afternoon']['end'], '19:00')
        self.assertEqual(form_data['today']['start'], '09:00')
        self.assertEqual(form_data['today']['end'], '19:00')
        self.assertEqual(form_data['default_locktime'], 2000)
        self.assertEqual(form_data['default_usage_count'], 5)



from cone.app import get_root
from cone.sql import get_session
from cone.sql import testing as sql_testing
from cone.tokens.browser.token import TokenAddForm
from cone.tokens.browser.token import TokenContent
from cone.tokens.browser.token import TokenEditForm
from cone.tokens.browser.token import TokenForm
from cone.tokens.browser.token import b64_qr_code
from cone.tokens.browser.token import qr_code
from cone.tokens.model import TokenNode
from cone.tokens.model import TokenRecord
from cone.tokens.settings import tokens_config
from cone.tokens.tests import tokens_layer
from cone.ugm import testing
from cone.ugm.testing import principals
from datetime import datetime 
from node.tests import NodeTestCase
from node.utils import UNSET
from pyramid.view import render_view_to_response
from yafowil.base import ExtractionError
import os
import uuid


class TestTokenViews(NodeTestCase):
    layer = tokens_layer

    def test_qr_code(self):
        data = qr_code('A')
        self.assertTrue(data.startswith(b'\x89PNG'))

    def test_b64_qr_code(self):
        data = b64_qr_code('A')
        self.assertTrue(data.startswith('data:image/png;base64'))

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    def test_download_qr_code(self):
        tokens = get_root()['tokens']
        token = TokenNode()
        tokens[str(token.uuid)] = token
        token.attrs['value'] = 'value'

        request = self.layer.new_request()
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'qr_code')
        self.assertTrue(res.body.startswith(b'\x89PNG'))
        self.assertEqual(res.content_type, 'image/png')
        self.assertEqual(
            res.content_disposition,
            f'attachment;filename={token.name}.png'
        )

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    def test_TokenContent(self):
        request = self.layer.new_request()
        token_tile = TokenContent()
        token_tile.model = TokenNode()
        token_tile.request = request
        # XXX


class TestTokenForms(NodeTestCase):
    layer = tokens_layer

    @testing.temp_directory
    def test_TokenForm(self, tempdir):
        tokens_config.config_file = os.path.join(tempdir, 'tokens.json')
        request = self.layer.new_request()

        # create token
        tokens = get_root()['tokens']
        token = TokenNode(parent=tokens)

        class TestTokenForm(TokenForm):
            def next(self, request):
                ...

        # prepare token form
        form_tile = TestTokenForm(attribute='render')
        form_tile.model = token
        form_tile.request = request
        form_tile.action_resource = 'tokenform'
        form_tile.prepare()

        # token form structure
        self.assertEqual(form_tile.form.name, 'tokenform')
        self.checkOutput("""
        <class 'yafowil.base.Widget'>: tokenform
        __<class 'yafowil.base.Widget'>: value
        __<class 'yafowil.base.Widget'>: valid_from
        __<class 'yafowil.base.Widget'>: valid_to
        __<class 'yafowil.base.Widget'>: usage_count
        __<class 'yafowil.base.Widget'>: lock_time
        __<class 'yafowil.base.Widget'>: save
        __<class 'yafowil.base.Widget'>: cancel
        """, form_tile.form.treerepr(prefix='_'))

        # empty extraction
        data = form_tile.form.extract(request=request)
        self.assertEqual(data.fetch('tokenform.value').extracted, UNSET)
        self.assertEqual(data.fetch('tokenform.valid_from').extracted, None)
        self.assertEqual(data.fetch('tokenform.valid_to').extracted, None)
        self.assertEqual(data.fetch('tokenform.usage_count').extracted, UNSET)
        self.assertEqual(data.fetch('tokenform.lock_time').extracted, UNSET)

        # extraction with empty values
        request.params['tokenform.value'] = ''
        request.params['tokenform.valid_from'] = ''
        request.params['tokenform.valid_from.time'] = ''
        request.params['tokenform.valid_to'] = ''
        request.params['tokenform.valid_to.time'] = ''
        request.params['tokenform.usage_count'] = ''
        request.params['tokenform.lock_time'] = ''
        request.params['action.tokenform.save'] = ''

        data = form_tile.form.extract(request=request)
        self.assertEqual(data.fetch('tokenform.value').extracted, '')
        self.assertEqual(data.fetch('tokenform.valid_from').extracted, None)
        self.assertEqual(data.fetch('tokenform.valid_to').extracted, None)
        self.assertEqual(data.fetch('tokenform.usage_count').extracted, 0)
        self.assertEqual(data.fetch('tokenform.lock_time').extracted, 0)

        # extraction with values
        request.params['tokenform.value'] = 'token value'
        request.params['tokenform.valid_from'] = '21.9.2023'
        request.params['tokenform.valid_from.time'] = '10:00'
        request.params['tokenform.valid_to'] = '21.9.2023'
        request.params['tokenform.valid_to.time'] = '16:00'
        request.params['tokenform.usage_count'] = '1'
        request.params['tokenform.lock_time'] = '100'

        data = form_tile.form.extract(request=request)
        self.assertEqual(data.fetch('tokenform.value').extracted, 'token value')
        self.assertEqual(
            data.fetch('tokenform.valid_from').extracted,
            datetime(2023, 9, 21, 10, 0)
        )
        self.assertEqual(
            data.fetch('tokenform.valid_to').extracted,
            datetime(2023, 9, 21, 16, 0)
        )
        self.assertEqual(data.fetch('tokenform.usage_count').extracted, 1)
        self.assertEqual(data.fetch('tokenform.lock_time').extracted, 100)

        # value validation
        session = get_session(request)

        token = TokenRecord()
        token.uid = uuid.uuid4()
        token.value = 'value'
        session.add(token)
        session.commit()

        request.params['tokenform.value'] = 'value'
        data = form_tile.form.extract(request=request)
        self.assertEqual(
            data.fetch('tokenform.value').errors,
            [ExtractionError('value_already_used')]
        )
        request.params['tokenform.value'] = 'token value'

        # time range validation
        request.params['tokenform.valid_from'] = '21.9.2023'
        request.params['tokenform.valid_from.time'] = '10:00'
        request.params['tokenform.valid_to'] = ''
        request.params['tokenform.valid_to.time'] = ''

        data = form_tile.form.extract(request=request)
        self.assertEqual(
            data.fetch('tokenform.valid_from').extracted,
            datetime(2023, 9, 21, 10, 0)
        )
        self.assertEqual(data.fetch('tokenform.valid_to').extracted, None)

        request.params['tokenform.valid_from'] = ''
        request.params['tokenform.valid_from.time'] = ''
        request.params['tokenform.valid_to'] = '21.9.2023'
        request.params['tokenform.valid_to.time'] = '16:00'

        data = form_tile.form.extract(request=request)
        self.assertEqual(data.fetch('tokenform.valid_from').extracted, None)
        self.assertEqual(
            data.fetch('tokenform.valid_to').extracted,
            datetime(2023, 9, 21, 16, 0)
        )

        request.params['tokenform.valid_from'] = '22.9.2023'
        request.params['tokenform.valid_from.time'] = '00:00'
        request.params['tokenform.valid_to'] = '21.9.2023'
        request.params['tokenform.valid_to.time'] = '00:00'

        data = form_tile.form.extract(request=request)
        self.assertEqual(data.fetch('tokenform.valid_from').errors, [])
        self.assertEqual(
            data.fetch('tokenform.valid_to').errors,
            [ExtractionError('timerange_error')]
        )

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    @sql_testing.delete_table_records(TokenRecord)
    @testing.temp_directory
    def test_TokenAddForm(self, tempdir):
        tokens_config.config_file = os.path.join(tempdir, 'tokens.json')
        request = self.layer.new_request()

        # create token
        tokens = get_root()['tokens']
        token = TokenNode(parent=tokens)

        # prepare token form
        form_tile = TokenAddForm(attribute='render')
        form_tile.model = token
        form_tile.request = request
        form_tile.action_resource = 'tokenform'
        form_tile.prepare()

        # prepare request, token gets created with default values
        request.params['tokenform.value'] = ''
        request.params['tokenform.valid_from'] = ''
        request.params['tokenform.valid_from.time'] = ''
        request.params['tokenform.valid_to'] = ''
        request.params['tokenform.valid_to.time'] = ''
        request.params['tokenform.usage_count'] = ''
        request.params['tokenform.lock_time'] = ''
        request.params['action.tokenform.save'] = '1'

        # save token
        with self.layer.authenticated('admin'):
            form_tile(token, request)

        # check if token has been added
        self.assertEqual(len(tokens), 1)
        token = tokens[token.name]
        self.assertEqual(token.attrs['value'], token.name)
        self.assertEqual(token.attrs['valid_from'], None)
        self.assertEqual(token.attrs['valid_to'], None)
        self.assertEqual(token.attrs['usage_count'], 0)
        self.assertEqual(token.attrs['lock_time'], 0)

        # create another token with custom values
        tokens.clear()
        tokens()
        token = TokenNode(parent=tokens)

        request.params['tokenform.value'] = 'token value'
        request.params['tokenform.valid_from'] = '21.9.2023'
        request.params['tokenform.valid_from.time'] = '10:00'
        request.params['tokenform.valid_to'] = '22.9.2023'
        request.params['tokenform.valid_to.time'] = '12:00'
        request.params['tokenform.usage_count'] = '10'
        request.params['tokenform.lock_time'] = '100'
        request.params['action.tokenform.save'] = '1'

        # save token
        with self.layer.authenticated('admin'):
            form_tile(token, request)

        self.assertEqual(len(tokens), 1)
        token = tokens[token.name]
        self.assertEqual(token.attrs['value'], 'token value')
        self.assertEqual(token.attrs['valid_from'], datetime(2023, 9, 21, 10, 0))
        self.assertEqual(token.attrs['valid_to'], datetime(2023, 9, 22, 12, 0))
        self.assertEqual(token.attrs['usage_count'], 10)
        self.assertEqual(token.attrs['lock_time'], 100)

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    @sql_testing.delete_table_records(TokenRecord)
    @testing.temp_directory
    def test_TokenEditForm(self, tempdir):
        tokens_config.config_file = os.path.join(tempdir, 'tokens.json')
        request = self.layer.new_request()

        # create token
        tokens = get_root()['tokens']
        token = TokenNode()
        tokens[str(token.uuid)] = token
        token.attrs['value'] = 'token value'
        token.attrs['valid_from'] = datetime(2023, 9, 21, 10, 0)
        token.attrs['valid_to'] = datetime(2023, 9, 22, 12, 0)
        token.attrs['usage_count'] = 10
        token.attrs['lock_time'] = 100

        # prepare token form
        form_tile = TokenEditForm(attribute='render')
        form_tile.model = token
        form_tile.request = request
        form_tile.action_resource = 'tokenform'
        form_tile.prepare()

        # prepare request
        request.params['tokenform.value'] = ''
        request.params['tokenform.valid_from'] = '21.9.2023'
        request.params['tokenform.valid_from.time'] = '08:00'
        request.params['tokenform.valid_to'] = '21.9.2023'
        request.params['tokenform.valid_to.time'] = '18:00'
        request.params['tokenform.usage_count'] = '-1'
        request.params['tokenform.lock_time'] = '0'
        request.params['action.tokenform.save'] = '1'

        # save token
        with self.layer.authenticated('admin'):
            form_tile(token, request)

        # check token has been edited
        self.assertEqual(token.attrs['value'], token.name)
        self.assertEqual(token.attrs['valid_from'], datetime(2023, 9, 21, 8, 0))
        self.assertEqual(token.attrs['valid_to'], datetime(2023, 9, 21, 18, 0))
        self.assertEqual(token.attrs['usage_count'], -1)
        self.assertEqual(token.attrs['lock_time'], 0)



import os
import webresource as wr


resources_dir = os.path.join(os.path.dirname(__file__), 'static')
cone_tokens_resources = wr.ResourceGroup(
    name='cone.tokens-tokens',
    directory=resources_dir,
    path='tokens'
)
cone_tokens_resources.add(wr.ScriptResource(
    name='cone-tokens-js',
    depends='cone-app-protected-js',
    resource='cone.tokens.js',
    compressed='cone.tokens.min.js'
))
cone_tokens_resources.add(wr.StyleResource(
    name='cone-tokens-css',
    resource='cone.tokens.css',
    compressed='cone.tokens.min.css'
))


def configure_resources(config, settings):
    config.register_resource(cone_tokens_resources)
    config.set_resource_include('cone-tokens-js', 'authenticated')
    config.set_resource_include('cone-tokens-css', 'authenticated')



from cone.app import import_from_string
from cone.app import main_hook
from cone.app import register_config
from cone.tokens.browser import configure_resources
from cone.tokens.settings import TokenSettings
from cone.tokens.settings import tokens_config
import cone.app
import logging


logger = logging.getLogger('cone.tokens')


@main_hook
def initialize_tokens(config, global_config, settings):
    # application startup initialization

    tokens_config.config_file = settings['cone.tokens.config_file']
    tokens_config.settings_node_path = settings.get(
        'cone.tokens.settings_node_path',
        'settings/token_settings'
    ).split('/')

    # register entry node
    tokens_entry_factory = settings.get(
        'cone.tokens.entryfactory',
        'cone.tokens.model.TokenContainer'
    )
    cone.app.register_entry('tokens', import_from_string(tokens_entry_factory))

    # settings
    register_config('token_settings', TokenSettings)

    # static resources
    configure_resources(config, settings)

    # add translation
    config.add_translation_dirs('cone.tokens:locale/')

    # scan browser package
    config.scan('cone.tokens.browser')



from cone.app.browser.authoring import ContentEditForm
from cone.app.browser.form import Form
from cone.app.browser.layout import ProtectedContentTile
from cone.app.browser.settings import SettingsBehavior
from cone.app.browser.utils import make_url
from cone.tile import tile
from cone.tokens.settings import TokenSettings
from cone.tokens.settings import tokens_config
from plumber import plumbing
from pyramid.i18n import TranslationStringFactory
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.compound import compound_extractor
from yafowil.persistence import node_attribute_writer
import datetime
import json


_ = TranslationStringFactory('cone.ugm')


@tile(
    name='content',
    path='templates/token_settings.pt',
    interface=TokenSettings,
    permission='manage')
class TokenSettingsContent(ProtectedContentTile):
    pass


@tile(name='editform', interface=TokenSettings, permission='edit')
@plumbing(SettingsBehavior, ContentEditForm)
class TokenSettingsForm(Form):
    action_resource = u'edit'
    form_name = 'tokensettingsform'
    show_contextmenu = False
    form_heading = _('token_settings', default='Token Settings')

    @property
    def message_factory(self):
        return _

    def timerange_extractor(self, widget, data):
        start = data.extracted['start']
        end = data.extracted['end']
        if start:
            s = start.split(':')
            s_h = int(s[0])
            s_m = int(s[1])
        if end:
            e = end.split(':')
            e_h = int(e[0])
            e_m = int(e[1])
        if start and end:
            now = datetime.datetime.now()
            dt_st = now.replace(hour=s_h, minute=s_m, second=0, microsecond=0)
            dt_end = now.replace(hour=e_h, minute=e_m, second=0, microsecond=0)
            if dt_st > dt_end:
                raise ExtractionError(_(
                    'end_time_before_start_time',
                    default='Start Time must be before End Time.'
                ))
        return {
            'start': start,
            'end': end
        }

    def prepare(self):
        action = make_url(
            self.request,
            node=self.model,
            resource=self.action_resource
        )
        form = self.form = factory('#form', name=self.form_name, props={
            'action': action,
            'persist_writer': node_attribute_writer
        })
        form['morning'] = factory(
            '#field:error:*morning:compound',
            props={
                'label': _('morning', default='Morning')
            },
            custom = {
                'morning': {
                    'extractors': [compound_extractor, self.timerange_extractor]
                }
            })
        start_required = _('start_time_required', default='Start Time is required.')
        end_required = _('end_time_required', default='End Time is required.')
        form['morning']['start'] = factory(
            '#field:time',
            value=self.model.attrs['morning']['start'],
            props={
                'required': start_required,
                'label': _('start', default='Start'),
                'timepicker': True,
                'time': True,
                'persist': True
            })
        form['morning']['end'] = factory(
            '#field:time',
            value=self.model.attrs['morning']['end'],
            props={
                'required': end_required,
                'label': _('end', default='End'),
                'timepicker': True,
                'time': True,
                'persist': True
            })
        form['afternoon'] = factory(
            '#field:error:*afternoon:compound',
            props={
                'label': _('afternoon', default='Afternoon')
            },
            custom = {
                'afternoon': {
                    'extractors': [compound_extractor, self.timerange_extractor]
                }
            })
        form['afternoon']['start'] = factory(
            '#field:time',
            value=self.model.attrs['afternoon']['start'],
            props={
                'required': start_required,
                'label': _('start', default='Start'),
                'timepicker': True,
                'time': True,
                'persist': True
            })
        form['afternoon']['end'] = factory(
            '#field:time',
            value=self.model.attrs['afternoon']['end'],
            props={
                'required': end_required,
                'label': _('end', default='End'),
                'timepicker': True,
                'time': True,
                'persist': True
            })
        form['today'] = factory(
            '#field:error:*today:compound',
            props={
                'label': _('today', default='Today')
            },
            custom = {
                'today': {
                    'extractors': [compound_extractor, self.timerange_extractor]
                }
            })
        form['today']['start'] = factory(
            '#field:time',
            value=self.model.attrs['today']['start'],
            props={
                'required': start_required,
                'label': _('start', default='Start'),
                'timepicker': True,
                'time': True,
                'persist': True
            })
        form['today']['end'] = factory(
            '#field:time',
            value=self.model.attrs['today']['end'],
            props={
                'required': end_required,
                'label': _('end', default='End'),
                'timepicker': True,
                'time': True,
                'persist': True
            })
        form['default_locktime'] = factory(
            '#field:number',
            value=self.model.attrs['default_locktime'],
            props={
                'datatype': int,
                'min': 0,
                'required': _('default_locktime_required', default='Default Locktime is required.'),
                'label': _('default_locktime', default='Default Locktime')
            })
        form['default_usage_count'] = factory(
            '#field:number',
            value=self.model.attrs['default_usage_count'],
            props={
                'datatype': int,
                'min': -1,
                'required': _('default_usage_count_required', default='Default number of uses required.'),
                'label': _('default_usage_count', default='Default Uses')
            })
        form['save'] = factory(
            'submit',
            props={
                'action': 'save',
                'expression': True,
                'handler': self.save,
                'next': self.next,
                'label': 'Save'
        })
        form['cancel'] = factory(
            'submit',
            props={
                'action': 'cancel',
                'expression': True,
                'skip': True,
                'next': self.next,
                'label': 'Cancel'
        })

    def save(self, widget, data):
        def fetch(name):
            return data.fetch(f'tokensettingsform.{name}').extracted
        with open(tokens_config.config_file, 'w') as f:
            json.dump({
                'morning': fetch('morning'),
                'afternoon': fetch('afternoon'),
                'today': fetch('today'),
                'default_locktime': fetch('default_locktime'),
                'default_usage_count': fetch('default_usage_count')
            }, f, indent=4)



from cone.app import get_root
from cone.sql import testing as sql_testing
from cone.tokens.api import TokenAPI
from cone.tokens.browser.api import read_datetime
from cone.tokens.browser.api import read_int
from cone.tokens.browser.api import read_string
from cone.tokens.exceptions import TokenException
from cone.tokens.exceptions import TokenValueError
from cone.tokens.model import TokenRecord
from cone.tokens.model import TokenUsageRecord
from cone.tokens.settings import tokens_config
from cone.tokens.tests import tokens_layer
from cone.ugm import testing
from cone.ugm.testing import principals
from datetime import datetime 
from node.tests import NodeTestCase
from pyramid.httpexceptions import HTTPForbidden
from pyramid.view import render_view_to_response
from unittest.mock import patch
import os
import uuid


class TestJSONAPI(NodeTestCase):
    layer = tokens_layer

    def test_read_string(self):
        request = self.layer.new_request()
        kw = dict()

        read_string(request, 'param', kw)
        self.assertEqual(kw, dict())

        request.params['param'] = ''
        read_string(request, 'param', kw)
        self.assertEqual(kw, dict(param=None))

        read_string(request, 'param', kw, default='default')
        self.assertEqual(kw, dict(param='default'))

        request.params['param'] = 'value'
        read_string(request, 'param', kw)
        self.assertEqual(kw, dict(param='value'))

    def test_read_datetime(self):
        request = self.layer.new_request()
        kw = dict()

        read_datetime(request, 'param', kw)
        self.assertEqual(kw, dict())

        request.params['param'] = ''
        read_datetime(request, 'param', kw)
        self.assertEqual(kw, dict(param=None))

        read_datetime(request, 'param', kw, default=datetime(2023, 9, 22))
        self.assertEqual(kw, dict(param=datetime(2023, 9, 22)))

        with self.assertRaises(TokenValueError) as arc:
            request.params['param'] = 'foo'
            read_datetime(request, 'param', kw)
        self.assertEqual(arc.exception.message, 'param: invalid datetime format')

        request.params['param'] = datetime(2023, 9, 22, 10, 30).isoformat()
        read_datetime(request, 'param', kw)
        self.assertEqual(kw, dict(param=datetime(2023, 9, 22, 10, 30)))

    def test_read_int(self):
        request = self.layer.new_request()
        kw = dict()

        read_int(request, 'param', kw)
        self.assertEqual(kw, dict())

        request.params['param'] = ''
        read_int(request, 'param', kw)
        self.assertEqual(kw, dict(param=0))

        read_int(request, 'param', kw, default=-1)
        self.assertEqual(kw, dict(param=-1))

        with self.assertRaises(TokenValueError) as arc:
            request.params['param'] = 'foo'
            read_int(request, 'param', kw)
        self.assertEqual(arc.exception.message, 'param: value is no integer')

        request.params['param'] = '1'
        read_int(request, 'param', kw)
        self.assertEqual(kw, dict(param=1))

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    @sql_testing.delete_table_records(TokenRecord)
    def test_query_token(self):
        tokens = get_root()['tokens']
        request = self.layer.new_request(type='json')

        with self.assertRaises(HTTPForbidden) as arc:
            render_view_to_response(tokens, request, 'query_token')
        self.assertEqual(
            str(arc.exception),
            'Unauthorized: query_token failed permission check'
        )

        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'query_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'Missing ``value`` parameter')

        request.params['value'] = 'token value'
        with patch.object(
            TokenAPI,
            'query_token',
            side_effect=Exception('Exception')
        ):
            with self.layer.authenticated('admin'):
                res = render_view_to_response(tokens, request, 'query_token')
            self.assertFalse(res.json['success'])
            self.assertEqual(res.json['message'], 'Exception')

        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'query_token')
        self.assertTrue(res.json['success'])
        self.assertEqual(res.json['token'], None)

        token_api = TokenAPI(request)
        token_uid = uuid.uuid4()
        token_api.add(token_uid, value='token value')

        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'query_token')
        self.assertTrue(res.json['success'])
        self.assertTrue(res.json['token'] is not None)
        self.assertEqual(res.json['token']['uid'], str(token_uid))

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    @sql_testing.delete_table_records(TokenRecord)
    @testing.temp_directory
    def test_add_token(self, tempdir):
        tokens_config.config_file = os.path.join(tempdir, 'tokens.json')
        tokens = get_root()['tokens']
        request = self.layer.new_request(type='json')
        request.method = 'POST'

        with self.assertRaises(HTTPForbidden) as arc:
            render_view_to_response(tokens, request, 'add_token')
        self.assertEqual(
            str(arc.exception),
            'Unauthorized: add_token failed permission check'
        )

        request.params['valid_from'] = 'foo'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'add_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'valid_from: invalid datetime format')
        del request.params['valid_from']

        request.params['valid_to'] = 'foo'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'add_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'valid_to: invalid datetime format')
        del request.params['valid_to']

        request.params['usage_count'] = 'foo'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'add_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'usage_count: value is no integer')
        del request.params['usage_count']

        request.params['lock_time'] = 'foo'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'add_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'lock_time: value is no integer')
        del request.params['lock_time']

        request.params['valid_from'] = datetime(2023, 9, 22).isoformat()
        request.params['valid_to'] = datetime(2023, 9, 21).isoformat()
        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'add_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(
            res.json['message'],
            'valid_from must be before valid_to'
        )
        del request.params['valid_from']
        del request.params['valid_to']

        with patch.object(
            TokenAPI,
            'add',
            side_effect=TokenException('TokenException')
        ):
            with self.layer.authenticated('admin'):
                res = render_view_to_response(tokens, request, 'add_token')
            self.assertFalse(res.json['success'])
            self.assertEqual(res.json['message'], 'TokenException')

        with patch.object(
            TokenAPI,
            'add',
            side_effect=Exception('Exception')
        ):
            with self.layer.authenticated('admin'):
                res = render_view_to_response(tokens, request, 'add_token')
            self.assertFalse(res.json['success'])
            self.assertEqual(res.json['message'], 'Exception')

        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'add_token')
        self.assertTrue(res.json['success'])

        token_uid = res.json['token_uid']
        token = tokens[token_uid]
        self.assertEqual(token.attrs['value'], token_uid)
        self.assertEqual(token.attrs['valid_from'], None)
        self.assertEqual(token.attrs['valid_to'], None)
        self.assertEqual(token.attrs['usage_count'], 0)
        self.assertEqual(token.attrs['lock_time'], 0)

        request.params['value'] = ''
        request.params['valid_from'] = ''
        request.params['valid_to'] = ''
        request.params['usage_count'] = ''
        request.params['lock_time'] = ''
        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'add_token')
        self.assertTrue(res.json['success'])

        token_uid = res.json['token_uid']
        token = tokens[token_uid]
        self.assertEqual(token.attrs['value'], token_uid)
        self.assertEqual(token.attrs['valid_from'], None)
        self.assertEqual(token.attrs['valid_to'], None)
        self.assertEqual(token.attrs['usage_count'], 0)
        self.assertEqual(token.attrs['lock_time'], 0)

        request.params['value'] = 'value'
        request.params['valid_from'] = datetime(2023, 9, 21).isoformat()
        request.params['valid_to'] = datetime(2023, 9, 22).isoformat()
        request.params['usage_count'] = '100'
        request.params['lock_time'] = '200'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(tokens, request, 'add_token')
        self.assertTrue(res.json['success'])

        token_uid = res.json['token_uid']
        token = tokens[token_uid]
        self.assertEqual(token.attrs['value'], 'value')
        self.assertEqual(token.attrs['valid_from'], datetime(2023, 9, 21))
        self.assertEqual(token.attrs['valid_to'], datetime(2023, 9, 22))
        self.assertEqual(token.attrs['usage_count'], 100)
        self.assertEqual(token.attrs['lock_time'], 200)

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    @sql_testing.delete_table_records(TokenRecord)
    def test_update_token(self):
        request = self.layer.new_request(type='json')
        request.method = 'POST'

        token_api = TokenAPI(request)
        token_uid = uuid.uuid4()
        token_api.add(token_uid)

        tokens = get_root()['tokens']
        token = tokens[str(token_uid)]

        with self.assertRaises(HTTPForbidden) as arc:
            render_view_to_response(token, request, 'update_token')
        self.assertEqual(
            str(arc.exception),
            'Unauthorized: update_token failed permission check'
        )

        request.params['valid_from'] = 'foo'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'update_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'valid_from: invalid datetime format')
        del request.params['valid_from']

        request.params['valid_to'] = 'foo'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'update_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'valid_to: invalid datetime format')
        del request.params['valid_to']

        request.params['usage_count'] = 'foo'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'update_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'usage_count: value is no integer')
        del request.params['usage_count']

        request.params['lock_time'] = 'foo'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'update_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(res.json['message'], 'lock_time: value is no integer')
        del request.params['lock_time']

        request.params['valid_from'] = datetime(2023, 9, 22).isoformat()
        request.params['valid_to'] = datetime(2023, 9, 21).isoformat()
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'update_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(
            res.json['message'],
            'valid_from must be before valid_to'
        )
        del request.params['valid_from']
        del request.params['valid_to']

        with patch.object(
            TokenAPI,
            'update',
            side_effect=TokenException('TokenException')
        ):
            with self.layer.authenticated('admin'):
                res = render_view_to_response(token, request, 'update_token')
            self.assertFalse(res.json['success'])
            self.assertEqual(res.json['message'], 'TokenException')

        with patch.object(
            TokenAPI,
            'update',
            side_effect=Exception('Exception')
        ):
            with self.layer.authenticated('admin'):
                res = render_view_to_response(token, request, 'update_token')
            self.assertFalse(res.json['success'])
            self.assertEqual(res.json['message'], 'Exception')

        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'update_token')
        self.assertTrue(res.json['success'])
        self.assertEqual(token.attrs['value'], str(token_uid))
        self.assertEqual(token.attrs['valid_from'], None)
        self.assertEqual(token.attrs['valid_to'], None)
        self.assertEqual(token.attrs['usage_count'], 0)
        self.assertEqual(token.attrs['lock_time'], 0)

        request.params['value'] = 'value'
        request.params['valid_from'] = datetime(2023, 9, 21).isoformat()
        request.params['valid_to'] = datetime(2023, 9, 22).isoformat()
        request.params['usage_count'] = '100'
        request.params['lock_time'] = '200'
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'update_token')
        self.assertTrue(res.json['success'])
        self.assertEqual(token.attrs['value'], 'value')
        self.assertEqual(token.attrs['valid_from'], datetime(2023, 9, 21))
        self.assertEqual(token.attrs['valid_to'], datetime(2023, 9, 22))
        self.assertEqual(token.attrs['usage_count'], 100)
        self.assertEqual(token.attrs['lock_time'], 200)

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    @sql_testing.delete_table_records(TokenRecord)
    def test_delete_token(self):
        request = self.layer.new_request(type='json')
        request.method = 'POST'

        token_api = TokenAPI(request)
        token_uid = uuid.uuid4()
        token_api.add(token_uid)

        tokens = get_root()['tokens']
        token = tokens[str(token_uid)]

        with self.assertRaises(HTTPForbidden) as arc:
            render_view_to_response(token, request, 'delete_token')
        self.assertEqual(
            str(arc.exception),
            'Unauthorized: delete_token failed permission check'
        )

        with patch.object(
            TokenAPI,
            'delete',
            side_effect=Exception('Exception')
        ):
            with self.layer.authenticated('admin'):
                res = render_view_to_response(token, request, 'delete_token')
            self.assertFalse(res.json['success'])
            self.assertEqual(res.json['message'], 'Exception')

        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'delete_token')
        self.assertTrue(res.json['success'])
        self.assertEqual(tokens.values(), [])

    @principals(users={'admin': {}}, roles={'admin': ['manager']})
    @sql_testing.delete_table_records(TokenRecord)
    @sql_testing.delete_table_records(TokenUsageRecord)
    def test_consume_token(self):
        request = self.layer.new_request(type='json')
        request.method = 'GET'

        token_api = TokenAPI(request)
        token_uid = uuid.uuid4()
        token_api.add(token_uid)

        tokens = get_root()['tokens']
        token = tokens[str(token_uid)]

        with self.assertRaises(HTTPForbidden) as arc:
            render_view_to_response(token, request, 'consume_token')
        self.assertEqual(
            str(arc.exception),
            'Unauthorized: consume_token failed permission check'
        )

        with patch.object(
            TokenAPI,
            'consume',
            side_effect=TokenException('TokenException')
        ):
            with self.layer.authenticated('admin'):
                res = render_view_to_response(token, request, 'consume_token')
            self.assertFalse(res.json['success'])
            self.assertEqual(res.json['message'], 'TokenException')

        with patch.object(
            TokenAPI,
            'consume',
            side_effect=Exception('Exception')
        ):
            with self.layer.authenticated('admin'):
                res = render_view_to_response(token, request, 'consume_token')
            self.assertFalse(res.json['success'])
            self.assertEqual(res.json['message'], 'Exception')

        token.attrs['usage_count'] = -1
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'consume_token')
        self.assertTrue(res.json['success'])
        self.assertTrue(res.json['consumed'])

        token.attrs['usage_count'] = 0
        with self.layer.authenticated('admin'):
            res = render_view_to_response(token, request, 'consume_token')
        self.assertFalse(res.json['success'])
        self.assertEqual(
            res.json['message'],
            f'Token {str(token_uid)} usage count exceeded'
        )



from cone.app.model import BaseNode
from cone.app.model import Metadata
from node.utils import instance_property
from pyramid.i18n import TranslationStringFactory
import json
import logging
import os


_ = TranslationStringFactory('cone.tokens')


class TokensConfig:
    config_file = None
    settings_node_path = None


tokens_config = TokensConfig()


def get_settings_node(model):
    node = model.root
    for name in tokens_config.settings_node_path:
        node = node[name]
    return node


default_token_settings = {
    'morning': {'start': '08:00', 'end': '12:00'},
    'afternoon': {'start': '12:00', 'end': '18:00'},
    'today': {'start': '08:00', 'end': '18:00'},
    'default_locktime': 0,
    'default_usage_count': 0,
}


class TokenSettings(BaseNode):

    @property
    def attrs(self):
        config_file = tokens_config.config_file
        if not os.path.exists(config_file):
            msg = f'Configuration file {config_file} not exists. Create it.'
            logging.info(msg)
            with open(config_file, 'w') as f:
                json.dump(default_token_settings, f, indent=4)
            return default_token_settings
        with open(config_file) as f:
            data = json.load(f)
        return data

    @instance_property
    def metadata(self):
        metadata = Metadata()
        metadata.title = _('token_settings_node', default='Token Settings')
        metadata.description = _(
            'token_settings_node_description',
            default='Token definition settings'
        )
        return metadata



from cone.tokens.api import TokenAPI
from cone.tokens.exceptions import TokenException
from cone.tokens.exceptions import TokenValueError
from cone.tokens.model import TokenContainer
from cone.tokens.model import TokenNode
from cone.tokens.settings import get_settings_node
from pyramid.view import view_config
import dateutil.parser
import uuid


def read_string(request, param, kw, default=None):
    if param in request.params:
        value = request.params[param]
        if not value:
            value = default
        kw[param] = value


def read_datetime(request, param, kw, default=None):
    if param in request.params:
        value = request.params[param]
        if not value:
            kw[param] = default
        else:
            try:
                kw[param] = dateutil.parser.isoparse(value)
            except ValueError:
                raise TokenValueError(f'{param}: invalid datetime format')


def read_int(request, param, kw, default=0):
    if param in request.params:
        value = request.params[param]
        if not value:
            kw[param] = default
        else:
            try:
                kw[param] = int(request.params[param])
            except ValueError:
                raise TokenValueError(f'{param}: value is no integer')


def format_datetime(dt):
    if not dt:
        return None
    return dt.isoformat()


@view_config(
    name='query_token',
    request_method='GET',
    accept='application/json',
    renderer='json',
    context=TokenContainer,
    permission='view')
def query_token(model, request):
    api = TokenAPI(request)
    try:
        value = request.params['value']
        token = api.query_token(value)
        if not token:
            return dict(success=True, token=None)
        return dict(
            success=True,
            token=dict(
                uid=str(token.uid),
                value=token.value,
                last_used=format_datetime(token.last_used),
                valid_from=format_datetime(token.valid_from),
                valid_to=format_datetime(token.valid_to),
                usage_count=token.usage_count,
                lock_time=token.lock_time,
                creator=token.creator,
                created=format_datetime(token.created),
                modified=format_datetime(token.modified)
            )
        )
    except KeyError:
        return dict(success=False, message='Missing ``value`` parameter')
    except Exception as e:
        return dict(success=False, message=str(e))


@view_config(
    name='consume_token',
    request_method='GET',
    accept='application/json',
    renderer='json',
    context=TokenNode,
    permission='view')  # XXX: ``consume`` permission
def consume_token(model, request):
    api = TokenAPI(request)
    uid = uuid.UUID(model.name)
    try:
        consumed = api.consume(uid)
    except TokenException as e:
        return e.as_json()
    except Exception as e:
        return dict(success=False, message=str(e))
    return dict(success=True, consumed=consumed)


@view_config(
    name='add_token',
    request_method='POST',
    accept='application/json',
    renderer='json',
    context=TokenContainer,
    permission='add')
def add_token(model, request):
    settings = get_settings_node(model)
    api = TokenAPI(request)
    uid = uuid.uuid4()
    kw = dict()
    read_string(request, 'value', kw)
    try:
        read_datetime(request, 'valid_from', kw)
        read_datetime(request, 'valid_to', kw)
        read_int(
            request,
            'usage_count',
            kw,
            default=settings.attrs['default_usage_count']
        )
        read_int(
            request,
            'lock_time',
            kw,
            default=settings.attrs['default_locktime']
        )
    except TokenValueError as e:
        return e.as_json()
    try:
        api.add(uid, **kw)
    except TokenException as e:
        return e.as_json()
    except Exception as e:
        return dict(success=False, message=str(e))
    return dict(success=True, token_uid=str(uid))


@view_config(
    name='update_token',
    request_method='POST',
    accept='application/json',
    renderer='json',
    context=TokenNode,
    permission='edit')
def update_token(model, request):
    api = TokenAPI(request)
    uid = uuid.UUID(model.name)
    kw = dict()
    read_string(request, 'value', kw)
    try:
        read_datetime(request, 'valid_from', kw)
        read_datetime(request, 'valid_to', kw)
        read_int(request, 'usage_count', kw)
        read_int(request, 'lock_time', kw)
    except TokenValueError as e:
        return e.as_json()
    try:
        api.update(uid, **kw)
    except TokenException as e:
        return e.as_json()
    except Exception as e:
        return dict(success=False, message=str(e))
    return dict(success=True)


@view_config(
    name='delete_token',
    request_method='POST',
    accept='application/json',
    renderer='json',
    context=TokenNode,
    permission='delete')
def delete_token(model, request):
    api = TokenAPI(request)
    uid = uuid.UUID(model.name)
    try:
        api.delete(uid)
    except Exception as e:
        return dict(success=False, message=str(e))
    return dict(success=True)



from node.utils import UNSET


class TokenException(Exception):
    error_code = UNSET

    def __init__(self, message):
        self.message = message
        super().__init__(message)

    def as_json(self):
        return dict(success=False, message=self.message)


class TokenNotExists(TokenException):
    error_code = 1

    def __init__(self, token_uid):
        self.token_uid = token_uid
        super().__init__(f'Token {self.token_uid} not exists')


class TokenUsageCountExceeded(TokenException):
    error_code = 2

    def __init__(self, token_uid):
        self.token_uid = token_uid
        super().__init__(f'Token {self.token_uid} usage count exceeded')


class TokenLockTimeViolation(TokenException):
    error_code = 3

    def __init__(self, token_uid):
        self.token_uid = token_uid
        super().__init__(f'Token {self.token_uid} is locked')


class TokenTimeRangeViolation(TokenException):
    error_code = 4

    def __init__(self, token_uid):
        self.token_uid = token_uid
        super().__init__(f'Token {self.token_uid} out of time range')


class TokenValueError(TokenException):
    error_code = 5

    def __init__(self, message):
        super().__init__(message)


class TokenAPIError(TokenException):
    error_code = 6

    def __init__(self, message):
        super().__init__(message)



from cone.app import get_root
from cone.sql import testing as sql_testing
from cone.tokens.model import TokenContainer
from cone.tokens.model import TokenNode
from cone.tokens.model import TokenRecord
from cone.tokens.tests import tokens_layer
from datetime import datetime
from node.tests import NodeTestCase
import uuid


class TestModel(NodeTestCase):
    layer = tokens_layer

    @sql_testing.delete_table_records(TokenRecord)
    def test_TokenContainer(self):
        tokens = get_root()['tokens']

        self.assertEqual(tokens.record_class, TokenRecord)
        self.assertEqual(tokens.child_factory, TokenNode)
        self.assertEqual(
            tokens.uuid,
            uuid.UUID('c40ef458-832f-42e6-9add-2dda2afb8920')
        )

        self.assertEqual(tokens.metadata.title, 'token_container_title')
        self.assertEqual(
            tokens.metadata.description,
            'token_container_description'
        )

        self.assertTrue(tokens.properties.in_navtree)
        self.assertTrue(tokens.properties.action_up)
        #self.assertTrue(tokens.properties.action_sharing)
        self.assertTrue(tokens.properties.action_view)
        self.assertTrue(tokens.properties.action_list)

    @sql_testing.delete_table_records(TokenRecord)
    def test_TokenNode(self):
        tokens = get_root()['tokens']
        self.assertEqual(isinstance(tokens, TokenContainer), True)

        # add token to tokens container
        token = TokenNode()
        token.attrs['value'] = 'value'
        token.attrs['valid_from'] = datetime(2023, 9, 21)
        token.attrs['valid_to'] = datetime(2023, 9, 22)
        token.attrs['lock_time'] = 0
        token.attrs['usage_count'] = -1
        token.attrs['creator'] = 'admin'
        token.attrs['created'] = datetime(2023, 9, 21)
        token.attrs['modified'] = datetime(2023, 9, 21)

        tokens[str(token.uuid)] = token
        token()

        # check if token has been added
        self.assertEqual(len(tokens), 1)
        token = tokens.values()[0]
        self.assertTrue(isinstance(token, TokenNode))
        self.assertEqual(token.attrs['value'], 'value')
        self.assertEqual(token.attrs['valid_from'], datetime(2023, 9, 21))
        self.assertEqual(token.attrs['valid_to'], datetime(2023, 9, 22))
        self.assertEqual(token.attrs['lock_time'], 0)
        self.assertEqual(token.attrs['usage_count'], -1)
        self.assertEqual(token.attrs['creator'], 'admin')
        self.assertEqual(token.attrs['created'], datetime(2023, 9, 21))
        self.assertEqual(token.attrs['modified'], datetime(2023, 9, 21))

        # check metadata
        self.assertEqual(token.metadata.title, 'value')
        self.assertEqual(token.metadata.creator, 'admin')
        self.assertEqual(token.metadata.created, datetime(2023, 9, 21))
        self.assertEqual(token.metadata.modified, datetime(2023, 9, 21))

        # check properties
        self.assertTrue(token.properties.action_up)
        self.assertEqual(token.properties.action_up_tile, 'content')
        self.assertTrue(token.properties.action_edit)
        self.assertTrue(token.properties.action_view)



from base64 import b64encode
from cone.app.browser.authoring import ContentAddForm
from cone.app.browser.authoring import ContentEditForm
from cone.app.browser.form import Form
from cone.app.browser.layout import ProtectedContentTile
from cone.app.browser.utils import make_url
from cone.app.browser.utils import request_property
from cone.app.utils import add_creation_metadata
from cone.app.utils import update_creation_metadata
from cone.sql import get_session
from cone.tile import tile
from cone.tokens.model import TokenNode
from cone.tokens.model import TokenRecord
from cone.tokens.settings import get_settings_node
from datetime import datetime
from node.utils import UNSET
from plumber import plumbing
from pyramid.i18n import TranslationStringFactory
from pyramid.response import Response
from pyramid.view import view_config
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.persistence import node_attribute_writer
import io
import json
import qrcode


_ = TranslationStringFactory('cone.tokens')


def qr_code(value):
    img = qrcode.make(value)
    data = io.BytesIO()
    img.save(data, format='png')
    return data.getvalue()


def b64_qr_code(value):
    return 'data:image/png;base64,' + b64encode(qr_code(value)).decode('utf-8')


@view_config(name='qr_code', context=TokenNode, permission='view')
def download_qr_code(model, request):
    response = Response(
        content_type='image/png',
        content_disposition=f'attachment;filename={model.name}.png'
    )
    response.body = qr_code(model.attrs['value'])
    return response


@tile(
    name='content',
    path='templates/token.pt',
    interface=TokenNode,
    permission='view')
class TokenContent(ProtectedContentTile):

    @property
    def token_settings(self):
        attrs = get_settings_node(self.model).attrs
        return json.dumps({
            'base_url': self.nodeurl,
            'timeranges': attrs
        })

    @request_property
    def qrcode(self):
        return b64_qr_code(self.model.attrs['value'])

    @property
    def lock_time_seconds(self):
        return f"{self.model.attrs.get('lock_time')} sec"

    @property
    def is_active(self):
        # check if token is active / valid, does not check lock time
        attrs = self.model.attrs
        if attrs.get('usage_count') == 0:
            return False
        if attrs.get('valid_to') and datetime.now() > attrs.get('valid_to'):
            return False
        if attrs.get('valid_from') and datetime.now() < attrs.get('valid_from'):
            return False
        return True

    @property
    def active_label(self):
        return 'Active' if self.is_active else 'Inactive'

    @property
    def cssclass(self):
        return 'btn-success' if self.is_active else 'btn-danger'

    def format_date(self, value):
        if isinstance(value, datetime):
            return value.strftime('%d.%m.%Y, %H:%M:%S')
        return 'Unlimited'


class TokenForm(Form):
    form_name = 'tokenform'

    @property
    def form_action(self):
        return make_url(
            self.request,
            node=self.model,
            resource=self.action_resource
        )

    def timerange_extractor(self, widget, data):
        extracted = data.extracted
        if extracted is UNSET:
            return extracted
        valid_from = data.fetch('tokenform.valid_from').extracted
        if valid_from and extracted:
            if valid_from >= extracted:
                raise ExtractionError(_(
                    'timerange_error',
                    default='Valid from date must be before valid to date'
                ))
        return extracted

    def value_extractor(self, widget, data):
        extracted = data.extracted
        if not extracted:
            return extracted
        session = get_session(self.request)
        existing_value = session.query(TokenRecord) \
            .filter(TokenRecord.value == extracted) \
            .filter(TokenRecord.uid != self.model.record.uid) \
            .one_or_none()
        if existing_value:
            raise ExtractionError(_(
                'value_already_used',
                default='Value already used by another token'
            ))
        return extracted

    def prepare(self):
        form = self.form = factory(
            '#form',
            name=self.form_name,
            props={
                'action': self.form_action,
                'persist_writer': node_attribute_writer
            }
        )
        attrs = self.model.attrs
        form['value'] = factory(
            '#field:*value:text',
            value=attrs.get('value', UNSET),
            props={
                'label': _('value', default='Value'),
                'persist': False
            },
            custom={
                'value': {
                    'extractors': [self.value_extractor]
                }
            }
        )
        form['valid_from'] = factory(
            '#field:datetime',
            value=attrs.get('valid_from', UNSET),
            props={
                'label': _('valid_from', default='Valid from'),
                'datepicker': True,
                'timepicker': True,
                'time': True,
                'locale': 'de',
                'persist': True
            }
        )
        form['valid_to'] = factory(
            '#field:*valid_to:datetime',
            value=attrs.get('valid_to', UNSET),
            props={
                'label': _('valid_to', default='Valid to'),
                'datepicker': True,
                'timepicker': True,
                'time': True,
                'locale': 'de',
                'persist': True
            },
            custom={
                'valid_to': {
                    'extractors': [self.timerange_extractor]
                }
            }
        )
        settings = get_settings_node(self.model)
        usage_count = attrs['usage_count']
        if usage_count is None:
            usage_count = settings.attrs['default_usage_count']
        form['usage_count'] = factory(
            '#field:number',
            value=usage_count,
            props={
                'label': _('usage_count', default='Usage Count'),
                'datatype': int,
                'emptyvalue': 0
            }
        )
        lock_time = attrs['lock_time']
        if lock_time is None:
            lock_time = settings.attrs['default_locktime']
        form['lock_time'] = factory(
            '#field:number',
            value=lock_time,
            props={
                'label': _('lock_time', default='Lock Time'),
                'datatype': int,
                'emptyvalue': 0
            }
        )
        form['save'] = factory(
            'submit',
            props={
                'action': 'save',
                'expression': True,
                'handler': self.save,
                'next': self.next,
                'label': _('save', default='Save')
            }
        )
        form['cancel'] = factory(
            'submit',
            props={
                'action': 'cancel',
                'expression': True,
                'skip': True,
                'next': self.next,
                'label': _('cancel', default='Cancel')
            }
        )

    def save(self, widget, data):
        data.write(self.model)
        value = data.fetch('tokenform.value').extracted
        # token uid gets used as value if no value given
        if not value:
            value = str(self.model.uuid)
        self.model.attrs['value'] = value


@tile(name='addform', interface=TokenNode, permission='add')
@plumbing(ContentAddForm)
class TokenAddForm(TokenForm):

    def save(self, widget, data):
        super(TokenAddForm, self).save(widget, data)
        self.model.parent[str(self.model.uuid)] = self.model
        add_creation_metadata(self.request, self.model.attrs)
        self.model()


@tile(name='editform', interface=TokenNode, permission='edit')
@plumbing(ContentEditForm)
class TokenEditForm(TokenForm):

    def save(self, widget, data):
        super(TokenEditForm, self).save(widget, data)
        update_creation_metadata(self.request, self.model.attrs)
        self.model()



from cone import sql
from cone.sql import get_session
from cone.sql import use_tm
from cone.tokens.exceptions import TokenLockTimeViolation
from cone.tokens.exceptions import TokenNotExists
from cone.tokens.exceptions import TokenTimeRangeViolation
from cone.tokens.exceptions import TokenUsageCountExceeded
from cone.tokens.exceptions import TokenValueError
from cone.tokens.model import TokenUsageRecord
from cone.tokens.model import TokenRecord
from datetime import datetime
from datetime import timedelta
from node.utils import UNSET
from node.utils import instance_property
from pyramid.i18n import TranslationStringFactory
from pyramid.threadlocal import get_current_request
import uuid


_ = TranslationStringFactory('cone.tokens')


class TokenAPI(object):

    def __init__(self, request=None):
       self.request = request

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.session.close()

    @instance_property
    def session(self):
        if self.request is None:
            return sql.session_factory()
        return get_session(self.request)

    def flush_or_commit(self):
        if use_tm() and self.request is not None:
            self.session.flush() # pragma: no cover
        else:
            self.session.commit()

    def get_token(self, token_uid):
        session = self.session
        token = session\
            .query(TokenRecord)\
            .filter(TokenRecord.uid == token_uid)\
            .one_or_none()
        if not token:
            raise TokenNotExists(token_uid)
        return token

    def query_token(self, value):
        session = self.session
        return session\
            .query(TokenRecord)\
            .filter(TokenRecord.value == value)\
            .one_or_none()

    def token_values(self, token_uids):
        session = self.session
        res = session\
            .query(TokenRecord.value)\
            .filter(TokenRecord.uid.in_(token_uids))\
            .all()
        return [it[0] for it in res]

    @property
    def _authenticated_userid(self):
        request = get_current_request()
        if request:
            return request.authenticated_userid

    def _log_usage(self, token_uid, error_code=None):
        record = TokenUsageRecord()
        record.uid = uuid.uuid4()
        record.token_uid = token_uid
        record.timestamp = datetime.now()
        record.error_code = error_code
        record.user = self._authenticated_userid
        self.session.add(record)
        self.flush_or_commit()

    def consume(self, token_uid):
        token = self.get_token(token_uid)
        if token.usage_count == 0:
            exc = TokenUsageCountExceeded(token_uid)
            self._log_usage(token_uid, error_code=exc.error_code)
            raise exc
        now = datetime.now()
        if token.last_used:
            if token.last_used + timedelta(0, token.lock_time) > now:
                exc = TokenLockTimeViolation(token_uid)
                self._log_usage(token_uid, error_code=exc.error_code)
                raise exc
        valid_from = token.valid_from
        valid_to = token.valid_to
        if valid_from and now < valid_from:
            exc = TokenTimeRangeViolation(token_uid)
            self._log_usage(token_uid, error_code=exc.error_code)
            raise exc
        if valid_to and now > valid_to:
            exc = TokenTimeRangeViolation(token_uid)
            self._log_usage(token_uid, error_code=exc.error_code)
            raise exc
        if token.usage_count != -1:
            token.usage_count -= 1
        token.last_used = now
        self._log_usage(token_uid)
        return True

    def add(
        self,
        token_uid,
        value=None,
        valid_from=None,
        valid_to=None,
        usage_count=0,
        lock_time=0
    ):
        try:
            self.get_token(token_uid)
            raise TokenValueError(f'Token with uid {token_uid} already exists')
        except TokenNotExists:
            if valid_from and valid_to and valid_from >= valid_to:
                raise TokenValueError('valid_from must be before valid_to')
            if not value:
                value = str(token_uid)
            session = self.session
            token = TokenRecord()
            token.uid = token_uid
            token.value = value
            token.valid_from = valid_from
            token.valid_to = valid_to
            token.lock_time = lock_time
            token.usage_count = usage_count
            token.creator = self._authenticated_userid
            now = datetime.now()
            token.created = now
            token.modified = now
            session.add(token)
            self.flush_or_commit()

    def update(
        self,
        token_uid,
        value=UNSET,
        valid_from=UNSET,
        valid_to=UNSET,
        usage_count=UNSET,
        lock_time=UNSET
    ):
        token = self.get_token(token_uid)
        session = self.session
        if value is not UNSET and value != token.value:
            existing = self.query_token(value)
            if existing and existing.uid != token_uid:
                raise TokenValueError('Given value already used by another token')
            token.value = value
        if valid_from is not UNSET:
            token.valid_from = valid_from
        if valid_to is not UNSET:
            token.valid_to = valid_to
        if lock_time is not UNSET:
            token.lock_time = lock_time
        if usage_count is not UNSET:
            token.usage_count = usage_count
        if token.valid_from and token.valid_to and token.valid_from >= token.valid_to:
            session.rollback()
            raise TokenValueError('valid_from must be before valid_to')
        token.modified = datetime.now()
        self.flush_or_commit()

    def delete(self, token_uid):
        session = self.session
        token = self.get_token(token_uid)
        session.delete(token)
        self.flush_or_commit()



from cone.app.interfaces import INavigationLeaf
from cone.app.model import Metadata
from cone.app.model import Properties
from cone.app.model import UUIDAttributeAware
from cone.app.model import node_info
from cone.sql import SQLBase
#from cone.sql.acl import SQLPrincipalACL
from cone.sql.model import GUID
from cone.sql.model import SQLRowNode
from cone.sql.model import SQLTableNode
from node.interfaces import IUUID
from node.utils import instance_property
from plumber import plumbing
from pyramid.i18n import TranslationStringFactory
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String
from zope.interface import implementer
import uuid


_ = TranslationStringFactory('cone.tokens')


class TokenRecord(SQLBase):
    __tablename__ = 'tokens'

    uid = Column(GUID, primary_key=True)
    value = Column(String)
    last_used = Column(DateTime)
    valid_from = Column(DateTime)
    valid_to = Column(DateTime)
    usage_count = Column(Integer)
    lock_time = Column(Integer)  # in seconds
    creator = Column(String)
    created = Column(DateTime)
    modified = Column(DateTime)


@node_info(
    name='token_node',
    title=_('token_node_title', default='Token'),
    description=_('token_node_description', default='Token'))
@plumbing(UUIDAttributeAware)
@implementer(INavigationLeaf)
class TokenNode(SQLRowNode):
    record_class = TokenRecord
    uuid_attribute_name = 'uid'

    @property
    def properties(self):
        props = Properties()
        props.action_up = True
        props.action_up_tile = 'content'
        props.action_edit = True
        props.action_view = True
        props.action_delete = True
        return props

    @instance_property
    def metadata(self):
        md = Metadata()
        md.title = str(self.attrs.get('value'))
        md.creator = self.attrs.get('creator')
        md.created = self.attrs.get('created')
        md.modified = self.attrs.get('modified')
        return md


@node_info(
    name='token_container',
    title=_('token_container_title', default='Tokens'),
    description=_('token_container_description', default='Tokens'),
    addables=['token_node'])
#@plumbing(SQLPrincipalACL)
@implementer(IUUID)
class TokenContainer(SQLTableNode):
    record_class = TokenRecord
    child_factory = TokenNode
    uuid = uuid.UUID('c40ef458-832f-42e6-9add-2dda2afb8920')

    @property
    def properties(self):
        props = Properties()
        props.in_navtree = True
        props.action_up = True
        #props.action_sharing = True
        props.action_view = True
        props.action_list = True
        return props

    @property
    def metadata(self):
        md = Metadata()
        info = self.nodeinfo
        md.title = info.title
        md.description = info.description
        return md


class TokenUsageRecord(SQLBase):
    __tablename__ = 'token_usage'

    uid = Column(GUID, primary_key=True)
    token_uid = Column(GUID)
    timestamp = Column(DateTime)
    error_code = Column(Integer)
    user = Column(String)



from bda.intellidatetime import DateTimeConversionError
from bda.intellidatetime import convert
from cone.app.browser import render_main_template
from cone.app.browser.actions import LinkAction
from cone.app.browser.contextmenu import context_menu_item
from cone.app.browser.layout import ProtectedContentTile
from cone.app.browser.utils import make_url
from cone.app.browser.utils import request_property
from cone.sql import get_session
from cone.tile import tile
from cone.tokens.browser.token import b64_qr_code
from cone.tokens.model import TokenContainer
from cone.tokens.model import TokenRecord
from cone.tokens.settings import get_settings_node
from pyramid.i18n import TranslationStringFactory
from pyramid.view import view_config
import json


_ = TranslationStringFactory('cone.tokens')


@tile(
    name='content',
    path='templates/tokens.pt',
    interface=TokenContainer,
    permission='view')
class TokensContent(ProtectedContentTile):
    ...


@context_menu_item(group='contentviews', name='tokens_overview')
class TokensOverviewAction(LinkAction):
    text = _('tokens_overview', default='Overview')
    icon = 'glyphicon glyphicon-list-alt'
    action = 'tokens_overview:#content:inner'
    path = 'href'

    @property
    def href(self):
        return make_url(self.request, node=self.model, resource='tokens_overview')

    @property
    def display(self):
        return isinstance(self.model, TokenContainer)

    @property
    def selected(self):
        return self.action_scope == 'tokens_overview'


@view_config(name='tokens_overview', context=TokenContainer, permission='view')
def tokens_overview(model, request):
    return render_main_template(model, request, 'tokens_overview')


@tile(
    name='tokens_overview',
    path='templates/tokens_overview.pt',
    interface=TokenContainer,
    permission='view')
class TokensOverview(ProtectedContentTile):

    @request_property
    def token_settings(self):
        return json.dumps({
            'base_url': self.nodeurl
        })

    @request_property
    def start(self):
        try:
            return convert(self.request.params.get('start', ''), locale='de')
        except DateTimeConversionError:
            return None

    @request_property
    def end(self):
        try:
            return convert(self.request.params.get('end', ''), locale='de')
        except DateTimeConversionError:
            return None

    @property
    def tokens(self):
        start = self.start
        end = self.end
        session = get_session(self.request)
        query = session.query(TokenRecord)
        if start:
            query = query.filter(TokenRecord.created >= start)
        if end:
            query = query.filter(TokenRecord.created <= end)
        tokens = query.all()
        return tokens

    def qrcode(self, value):
        return b64_qr_code(value)

    def format_date(self, date, default=''):
        if not date:
            return default
        return date.strftime('%d.%m.%Y')
